@@ -1,6 +1,4 @@
-# -=- encoding: utf-8 -=-
-
-import json
+from socketio.defaultjson import default_json_dumps, default_json_loads
 
 MSG_TYPES = {
     'disconnect': 0,
@@ -34,12 +32,10 @@ socketio_packet_attributes = ['type', 'name', 'data', 'endpoint', 'args',
                               'ackId', 'reason', 'advice', 'qs', 'id']
 
 
-
-def encode(data):
+def encode(data, json_dumps=default_json_dumps):
     """
     Encode an attribute dict into a byte string.
     """
-    ack_with_data = True
     payload = ''
     msg = str(MSG_TYPES[data['type']])
 
@@ -63,17 +59,18 @@ def encode(data):
         if msg == '3':
             payload = data['data']
         if msg == '4':
-            payload = json.dumps(data['data'], separators=(',', ':'))
+            payload = json_dumps(data['data'])
         if msg == '5':
             d = {}
             d['name'] = data['name']
             if 'args' in data and data['args'] != []:
                 d['args'] = data['args']
-            payload = json.dumps(d, separators=(',', ':'))
+            payload = json_dumps(d)
         if 'id' in data:
             msg += ':' + str(data['id'])
-            if ack_with_data:
-                msg += '+:'
+            if data['ack'] == 'data':
+                msg += '+'
+            msg += ':'
         else:
             msg += '::'
         if 'endpoint' not in data:
@@ -85,9 +82,9 @@ def encode(data):
 
     elif msg == '6':
         # '6:::' [id] '+' [data]
-        msg += ':::' + str(data['ackId'])
+        msg += '::' + data.get('endpoint', '') + ':' + str(data['ackId'])
         if 'args' in data and data['args'] != []:
-            msg += '+' + json.dumps(data['args'], separators=(',', ':'))
+            msg += '+' + json_dumps(data['args'])
 
     elif msg == '7':
         # '7::' [endpoint] ':' [reason] '+' [advice]
@@ -98,10 +95,14 @@ def encode(data):
             msg += '+' + str(ERROR_ADVICES[data['advice']])
         msg += data['endpoint']
 
+    # NoOp, used to close a poll after the polling duration time
+    elif msg == '8':
+        msg += '::'
+
     return msg
 
 
-def decode(rawstr):
+def decode(rawstr, json_loads=default_json_loads):
     """
     Decode a rawstr packet arriving from the socket into a dict.
     """
@@ -123,7 +124,12 @@ def decode(rawstr):
             decoded_msg['ack'] = True
 
     # common to every message
-    decoded_msg['type'] = MSG_VALUES[int(msg_type)]
+    msg_type_id = int(msg_type)
+    if msg_type_id in MSG_VALUES:
+        decoded_msg['type'] = MSG_VALUES[int(msg_type)]
+    else:
+        raise Exception("Unknown message type: %s" % msg_type)
+
     decoded_msg['endpoint'] = endpoint
 
     if len(split_data) > 3:
@@ -142,25 +148,26 @@ def decode(rawstr):
         decoded_msg['data'] = data
 
     elif msg_type == "4":  # json msg
-        decoded_msg['data'] = json.loads(data)
+        decoded_msg['data'] = json_loads(data)
 
     elif msg_type == "5":  # event
         try:
-            data = json.loads(data)
+            data = json_loads(data)
         except ValueError, e:
             print("Invalid JSON event message", data)
-
-        decoded_msg['name'] = data.pop('name')
-        if 'args' in data:
-            decoded_msg['args'] = data['args']
-        else:
             decoded_msg['args'] = []
+        else:
+            decoded_msg['name'] = data.pop('name')
+            if 'args' in data:
+                decoded_msg['args'] = data['args']
+            else:
+                decoded_msg['args'] = []
 
     elif msg_type == "6":  # ack
         if '+' in data:
             ackId, data = data.split('+')
             decoded_msg['ackId'] = int(ackId)
-            decoded_msg['args'] = json.loads(data)
+            decoded_msg['args'] = json_loads(data)
         else:
             decoded_msg['ackId'] = int(data)
             decoded_msg['args'] = []
@@ -177,8 +184,6 @@ def decode(rawstr):
             else:
                 decoded_msg['reason'] = ''
     elif msg_type == "8":  # noop
-        return None
-    else:
-        raise Exception("Unknown message type: %s" % msg_type)
+        pass
 
     return decoded_msg
